package reliability;

import debug.*;
import timer.*;
import network.*;

import java.io.*;
import java.util.*;

/**
  Implementation of a stop-and-wait sender
  <p>
  Provided implementation shows stop-and-wait protocol,
  plus all the general setup and use of the SendBuffer
  <p>
  This implementation uses both the SendBuffer and
  Network classes.  You are not required to use
  these, but they simplify interaction with the timer and
  the underlying network.  If you want to modify or replace
  them, be sure to understand them first.
 */
public class SAWSender extends TimerClient
                               implements ReliableSender, PacketReceiver
{
  /**
    Create a new stop-and-wait sender.
    
    @param network the network to communicate on
    @param address the address of this host
    @param timeout value in usec
    */
  public SAWSender (Network network, Address address, long timeout)
  {
    this.address = address;
    destinationAddress = null;
    
    this.network = network;
    network.AddReceiver (this, address);
    
    buffer = new SendBuffer (50 * network.MTU ());
    
    // The following is for stop-and-wait, feel free to remove
    // and modify for sliding window protocol
    currentPacket = null;
    currentSequenceNum = 0;
    currentTimeout = null;
    timeoutValue = timeout;
  }
  
  private Network network;
  private Address address;
  private SendBuffer buffer;
  private Address destinationAddress;

  /**
    Open a connection to the ReliableReceiver represented by
    the address.  
    <p>
    For stop-and-wait, true connection setup is not implemented,
    and this jsut sets the destination
    
    @param destination the address of the receiver
    */
  public void Open (Address destination)
  {
    destinationAddress = destination;
  }
  
  /**
    Send a block of data on the connection.  Any amount of
    data may be sent.  This call may block if there is 
    insufficient buffer space.  Clients may make any number
    of <tt>Send</tt> calls per open connection.
    
    @param data the data (any amount) to send on the connection
    @throws ConnectionNotOpenException if the connection is
            not open
    */
   public void Send (byte[] data) throws ConnectionNotOpenException
   {
      if (destinationAddress == null)
         throw new ConnectionNotOpenException ();

      int previous = buffer.AppendBytes (data);

      // If the buffer had no data and the last packet sent has been 
      // acknowledged then send the next packet
      if (previous == 0 && currentPacket == null)  // Not currently sending
         SendNextPacket ();   // so start again
   }

  /**
    Close the current connection.      
    @throws ConnectionNotOpenException if the connection is
    not open
    */
  public void Close () throws ConnectionNotOpenException
  {
    if (destinationAddress == null)
      throw new ConnectionNotOpenException ();
    
      destinationAddress = null;
  }
  
  /**
    Processing that happens when we receive a packet
    <p>
    For stop-and-wait: send the next packet
    */
  public void PacketArrived (NetworkPacket pk)
  {
    ReliablePacket p = (ReliablePacket) pk;

    Debug.println('r',"Received ack for " + p.sequenceNum);

    if (p.sequenceNum == currentSequenceNum) { 
      if (currentTimeout != null)
	Timer.global.CancelTimeout (currentTimeout);
      currentTimeout = null; 
      currentPacket = null; 
      SendNextPacket ();
    }
  }

   /**
      Processing that happens when we get a timeout
      <p>
      For stop-and-wait: resend the current packet
    */
   public void Timeout (Object info)
   {
     Debug.println('r', "Got a Timeout at sender " + (currentSequenceNum)); 
     SendCurrentPacket ();
   }

  private ReliablePacket currentPacket;
  private int currentSequenceNum;
  private TimeoutID currentTimeout;
  private long timeoutValue; // in usec
  
  /**
    Send the next packet 
    */
  private void SendNextPacket ()
  {
    byte[] data = buffer.RemoveBytes (network.MTU ());
    
    if (data.length > 0)
      {
	Debug.println('r', "Sending packet number " + (currentSequenceNum+1)); 

	currentPacket = new ReliablePacket ();
	
	currentPacket.sourceAddr = address;
	currentPacket.destinationAddr = destinationAddress;
	
	++currentSequenceNum;
	currentPacket.sequenceNum = currentSequenceNum;
	currentPacket.ack = false;

	currentPacket.data = data;
	
	network.TransmitPacket (currentPacket);
	currentTimeout = Timer.global.SetTimeout (this, timeoutValue,
						  Timer.US, null);
	
      }
  }

   /**
     Send the current packet 
     */
  private void SendCurrentPacket ()
  {
    Debug.println('r', "RESending packet number " + (currentSequenceNum)); 
    if (currentPacket != null)
      {
	network.TransmitPacket (currentPacket);
      }
    
    currentTimeout = Timer.global.SetTimeout (this, timeoutValue,
					      Timer.US, null);
  }
  
}




